__all__ = ['bind', 'BindingWrapper', 'Binding']

from collections.abc import Callable, Hashable, Iterable
from functools import update_wrapper
from typing import (
  TYPE_CHECKING,
  Any,
  Generic,
  NamedTuple,
  Protocol,
  TypeAlias,
  TypeVar,
  overload,
)

from pyriak import _SENTINEL
from pyriak.eventkey import key_functions


if TYPE_CHECKING:
  from pyriak.space import Space


_T = TypeVar('_T')
_R = TypeVar('_R')
_S = TypeVar('_S')

_Callback: TypeAlias = Callable[['Space', _T], _R]


_empty_frozenset: frozenset[object] = frozenset()


class Binding(NamedTuple):
  event_type: type
  priority: Any
  keys: frozenset[Hashable]


class BindingWrapper(Generic[_T, _R]):
  """A wrapper for the event handler callback which holds the bindings."""

  __wrapped__: _Callback[_T, _R]

  def __init__(self, wrapped: _Callback[_T, _R], bindings: tuple[Binding, ...], /):
    self.__bindings__ = bindings
    update_wrapper(self, wrapped)

  def __call__(self, space: 'Space', event: _T, /) -> _R:
    return self.__wrapped__(space, event)

  def __get__(self, obj, objtype=None):
    wrapped = self.__wrapped__
    try:
      descr_get = type(wrapped).__get__  # type: ignore[attr-defined]
    except AttributeError:
      return wrapped
    return descr_get(wrapped, obj, objtype)


class _Decorator(Protocol, Generic[_T]):
  @overload
  def __call__(
    self, callback: BindingWrapper[_S, _R], /
  ) -> BindingWrapper[_S | _T, _R]: ...
  @overload
  def __call__(self, callback: _Callback[_T, _R], /) -> BindingWrapper[_T, _R]: ...


@overload
def bind(
  event_type: type[_T], priority: Any, /
) -> _Decorator[_T]: ...
@overload
def bind(
  event_type: type[_T], priority: Any, /, *, key: Hashable
) -> _Decorator[_T]: ...
@overload
def bind(
  event_type: type[_T], priority: Any, /, *, keys: Iterable[Hashable]
) -> _Decorator[_T]: ...
def bind(event_type, priority, /, *, key=_SENTINEL, keys=_SENTINEL):
  """Bind a callback to an event type.


  """
  if not isinstance(event_type, type):
    raise TypeError(f'{event_type!r} is not a type')
  try:
    hash(event_type)
  except TypeError:
    raise TypeError(f'{event_type!r} is not hashable') from None
  if key is not _SENTINEL:
    if keys is not _SENTINEL:
      raise TypeError("bind() cannot be passed both 'key' and 'keys' kwargs")
    keys = frozenset([key])
  else:
    keys = frozenset(keys) if keys is not _SENTINEL else _empty_frozenset
  if keys and not key_functions.exists(event_type):
    raise ValueError(
      f'bind(): keys were provided but no key function exists for {event_type!r}'
    )
  def decorator(callback, /):
    if not isinstance(callback, BindingWrapper):
      return BindingWrapper(callback, (Binding(event_type, priority, keys),))
    for binding in callback.__bindings__:
      if event_type is binding.event_type:
        raise ValueError(
          f'event handler {callback.__wrapped__!r} already has binding for {event_type!r}'
        )
    callback.__bindings__ += (Binding(event_type, priority, keys),)
    return callback
  return decorator



# ----------------------------------------------------------------------------
# pyriak
# Copyright (c) 2024 aatle

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ----------------------------------------------------------------------------

"""Pyriak is a framework for ECS architecture.

This implementation integrates the concepts of event handling and polymorphism.
The package is written in pure python and has no dependencies.

This module is the top-level module of the package and contains
most of the things necessary to use this package.
"""

__all__ = [
  'bind',
  'Space',
  'Entity',
  'EntityId',
  'System',
  'EventQueue',
  'subclasses',
  'strict_subclasses',
  'key_functions',
  'set_key',
  'tagclass',
  'first',
]

from collections.abc import (
  Generator as _Generator,
  Hashable as _Hashable,
  MutableSequence as _MutableSequence,
)
from enum import Enum as _Enum
from typing import (
  Any as _Any,
  TypeAlias as _TypeAlias,
  TypeVar as _TypeVar,
)
from weakref import ref as _weakref


_TypeT = _TypeVar('_TypeT', bound=type)
_T = _TypeVar('_T')
_D = _TypeVar('_D')


System: _TypeAlias = _Hashable


EventQueue: _TypeAlias = _MutableSequence[object]


dead_weakref: _weakref[_Any] = _weakref(set())


_get_subclasses = type.__subclasses__

def subclasses(cls: _TypeT, /) -> _Generator[_TypeT, None, _TypeT]:
  """Generator of the class and all of the class's subclasses.

  Metaclasses also work.
  There may be duplicates in the case of multiple inheritance.
  Note: overridden implementations of __subclasses__ are ignored
  because type.__subclasses__ is used, which also prevents cycles.

  The order the classes are returned is not specified, however:
  It is guaranteed that the cls passed in is the first class yielded.
  For a tree inheritance structure (single inheritance only), a subclass
  will never be yielded before any of its superclasses.

  'Return' the cls passed in. (This value is accessible through
  the StopIteration that is raised).
  """
  yield cls
  get_subclasses = _get_subclasses
  stack = get_subclasses(cls)
  pop = stack.pop
  while stack:
    subclass = pop()
    yield subclass
    stack += get_subclasses(subclass)
  return cls


def strict_subclasses(cls: _TypeT, /) -> _Generator[_TypeT, None, _TypeT]:
  """Same as subclasses, but does not yield the original class."""
  get_subclasses = _get_subclasses
  stack = get_subclasses(cls)
  pop = stack.pop
  while stack:
    subclass = pop()
    yield subclass
    stack += get_subclasses(subclass)
  return cls


def tagclass(cls: type) -> type:
  namespace = dict(cls.__dict__)
  namespace.setdefault('__slots__', ())
  namespace.pop('__dict__', None)
  namespace.pop('__weakref__', None)
  qualname = getattr(cls, '__qualname__', None)
  cls = type(cls)(cls.__name__, cls.__bases__, namespace)
  if qualname is not None:
    cls.__qualname__ = qualname
  def __eq__(self, other):
    if other is self or type(other) is type(self):
      return True
    return NotImplemented
  def __hash__(self):
    return hash((type(self),))
  cls.__eq__ = __eq__  # type: ignore[method-assign, assignment]
  cls.__hash__ = __hash__  # type: ignore[method-assign, assignment]
  return cls


def first(arg: _T, *args: _Any) -> _T:  # noqa: ARG001
  return arg


class _Sentinel(_Enum):
  SENTINEL = 1

_SENTINEL: _Sentinel = _Sentinel.SENTINEL


# circular imports
from pyriak.bind import bind  # noqa: E402
from pyriak.entity import Entity, EntityId  # noqa: E402
from pyriak.eventkey import key_functions, set_key  # noqa: E402
from pyriak.managers.entitymanager import QueryResult as QueryResult  # noqa: E402
from pyriak.space import Space  # noqa: E402


# cleanup namespace
del _MutableSequence, _Enum, _TypeVar, _weakref



__all__ = [
  'ComponentAdded',
  'ComponentRemoved',
  'EntityAdded',
  'EntityRemoved',
  'SystemAdded',
  'SystemRemoved',
  'EventHandlerAdded',
  'EventHandlerRemoved',
  'StateAdded',
  'StateRemoved',
]

from collections.abc import Hashable, Iterator
from typing import TYPE_CHECKING, Any

from pyriak import subclasses as _subclasses
from pyriak.eventkey import set_key as _set_key


if TYPE_CHECKING:
  from pyriak import System
  from pyriak.bind import _Callback
  from pyriak.entity import Entity
  from pyriak.managers.systemmanager import _EventHandler


class EntityAdded:
  def __init__(self, entity: 'Entity'):
    self.entity = entity

class EntityRemoved:
  def __init__(self, entity: 'Entity'):
    self.entity = entity


def _component_type_key(event: 'ComponentAdded | ComponentRemoved') -> Iterator[type]:
  yield from type(event.component).__mro__

@_set_key(_component_type_key)
class ComponentAdded:
  def __init__(self, entity: 'Entity', component: object):
    self.entity = entity
    self.component = component

@_set_key(_component_type_key)
class ComponentRemoved:
  def __init__(self, entity: 'Entity', component: object):
    self.entity = entity
    self.component = component


def _system_key(event: 'SystemAdded | SystemRemoved') -> 'System':
  return event.system

@_set_key(_system_key)
class SystemAdded:
  def __init__(self, system: 'System'):
    self.system = system

@_set_key(_system_key)
class SystemRemoved:
  def __init__(self, system: 'System'):
    self.system = system


def _state_type_key(event: 'StateAdded | StateRemoved') -> Iterator[type]:
  yield from type(event.state).__mro__

@_set_key(_state_type_key)
class StateAdded:
  def __init__(self, state: object):
    self.state = state

@_set_key(_state_type_key)
class StateRemoved:
  def __init__(self, state: object):
    self.state = state


def _handler_key(event: 'EventHandlerAdded | EventHandlerRemoved') -> Iterator[type]:
  yield from _subclasses(event.event_type)

@_set_key(_handler_key)
class EventHandlerAdded:
  def __init__(
    self, _handler: '_EventHandler', _event_type: type, _keys: frozenset[Hashable]
  ):
    self._handler = _handler
    self.event_type = _event_type
    self.keys = _keys

  @property
  def system(self):
    return self._handler.system

  @property
  def callback(self):
    return self._handler.callback

  @property
  def name(self):
    return self._handler.name

  @property
  def priority(self):
    return self._handler.priority

  @property
  def key(self):
    [key] = self.keys
    return key

@_set_key(_handler_key)
class EventHandlerRemoved:
  def __init__(
    self, _system: 'System', _callback: '_Callback', _name: str, _priority: Any,
    _event_type: type, _keys: frozenset[Hashable]
  ):
    self.system = _system
    self.callback = _callback
    self.name = _name
    self.priority = _priority
    self.event_type = _event_type
    self.keys = _keys

  @property
  def key(self):
    [key] = self.keys
    return key
